def define_ToolTips(GUI):
    GUI.ChooseSeq_Button.setToolTip('Single Fasta file with training sequences. \n'
                                    'The header of each sequence must be a unique \n'
                                    'sequence identifier and each sequence must follow the \n'
                                    'IUPAC nomenclature code for nucleic acids.')

    GUI.ChooseGT_Button.setToolTip('Tab separated file with possible labels for the training dataset. \n'
    'This labels wont be used during trainig, just for testing purposes. \n'
    '(Check documentation for info on file headers)')

    GUI.input_k.setToolTip('k-mer length')

    GUI.input_n_mimics.setToolTip('Number of data augmentations per sequence \n' 
                                  'that will be considered during training')


    GUI.input_n_epochs.setToolTip('Number of training epochs. An epoch is defined \n' 
                                 'as a training iteration over all the training pairs.')

    GUI.input_n_clusters.setToolTip('Expected or maximum number of clusters to find. \n'
                                    'It should be equal or greater than n_true_clusters \n'
                                    'when GT is provided. \n'
                                    'NOTE: Use 0 for automatically finding fine-grained \n'
                                    'clusters')

    GUI.input_scheduler.setToolTip('Learning rate schedule to train the neural network.')

    GUI.input_batch_sz.setToolTip('Number of data pairs that the network will receive \n'
                                  'simultaneouly during training. A larger batch may \n'
                                  'improve convergence but it may harm the accuracy \n')

    GUI.input_lambda.setToolTip('Hyperparameter to control cluster balance. \n'
                                'Use lambda: 1.2 when unbalanced clusters are expected \n'
                                'Use lambda: 2.8 when perfectly balanced clusters are expected \n')

    GUI.input_weight.setToolTip('Relative importance of the contrastive objective on \n'
                                'the final loss. Use a higher value when low intra- \n'
                                'cluster distance is expected and a lower value when \n'
                                'high intra-cluster variability is expected')
                            

    GUI.Silhouette.setToolTip('Higher is better - This measure compares the cluster \n'
                            ' assignment of each sequence with the assignment \n'
                            'of the closest sequence assigned to a different cluster. \n'
                            'Range: [-1,1]')

    GUI.DB_Index.setToolTip('Lower is better - Measures the average distance between \n'
                            'clusters, relative to their sizes. Range: [0,2] \n')                         

    GUI.Submit_Button.setToolTip('Parse training Information')



import torch
import torch.nn as nn
from torch.utils.data import Dataset

class myNet(nn.Module):
    def __init__(self, n_input, n_output):
        super(myNet, self).__init__()
        self.n_input = n_input
        self.layers = nn.Sequential(
                nn.Linear(n_input, 400),
                nn.ReLU(),
                nn.Dropout(p=0.5),
                nn.Linear(400, 128),
                nn.LeakyReLU()
        )
        
        self.instance = nn.Linear(128, 64) 
        
        self.classifier = nn.Sequential(
                nn.Dropout(p=0.5),
                nn.Linear(128, n_output),  
                nn.Softmax(dim=1)
        )
            
    def forward(self, x):
        x = x.view(-1, self.n_input)
        x = self.layers(x)
        latent = self.instance(x)
        out = self.classifier(x)
        return out, latent
    
class NetLinear(nn.Module):
    def __init__(self, n_input, n_output):
        super(NetLinear, self).__init__()
        self.n_input = n_input
        self.layers  = nn.Sequential(

            nn.Linear(n_input, 512),
            nn.ReLU(),
            nn.Dropout(p=0.5),
            nn.Linear(512, 64)            
        )

        self.classifier = nn.Sequential(
            nn.ReLU(),
            nn.Dropout(p=0.5),
            nn.Linear(64, n_output),  
            nn.Softmax(dim=1)
        )

    def forward(self, x):
        x = x.view(-1, self.n_input)
        latent = self.layers(x)
        out = self.classifier(latent)
        return out, latent
    

class myDataset(Dataset):
    
    def __init__(self, data, labels=None, transform=None):
        
        if transform: 
            self.data=transform(data)           
        else:
            self.data = data
            
        self.labels = labels

        
    def __len__(self):
        return self.data.shape[0]
    
    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        
        return {'features': self.data[idx, :], 'labels': self.labels[idx]}
    
    
class Dummy_Net(nn.Module):
    def __init__(self, n_input, n_output):
        super(Dummy_Net, self).__init__()
        self.classifier = nn.Sequential(
                nn.Linear(n_input,64),
                nn.ReLU(),
                nn.Dropout(p=0.3),
                nn.Linear(64,128),
                nn.ReLU(),
                nn.Dropout(p=0.2),
                nn.Linear(128, n_output),
                nn.Softmax(dim=1)
        )
        
    def forward(self, x):
        x = x.view(-1, x.shape[2])
        return self.classifier(x)


import sys
import pyximport 
pyximport.install()

from .kmers import kmer_counts

import random, itertools
import numpy as np
import pandas as pd
import sklearn.metrics.cluster as metrics
from sklearn.cluster import KMeans

import torch 
from torch.utils.data import Dataset, DataLoader
from torch.utils.data import DataLoader

from scipy.optimize import linear_sum_assignment
from sklearn.preprocessing import StandardScaler

import matplotlib.pyplot as plt
from colorsys import hsv_to_rgb

def check_sequence(header, seq):
    """
    Adapted from VAMB: https://github.com/RasmussenLab/vamb
    
    Check that there're no invalid characters or bad format
    in the file. 
    
    Note: The GAPS ('-') that are introduced from alignment
    are considered valid characters. 
    """

    if len(header) > 0 and (header[0] in ('>', '#') or header[0].isspace()):
        raise ValueError('Bad character in sequence header')
    if '\t' in header:
        raise ValueError('tab included in header')

    basemask = bytearray.maketrans(b'acgtuUswkmyrbdhvnSWKMYRBDHV-',
                                   b'ACGTTTNNNNNNNNNNNNNNNNNNNNNN')

    masked = seq.translate(basemask, b' \t\n\r')
    stripped = masked.translate(None, b'ACGTN')
    if len(stripped) > 0:
        bad_character = chr(stripped[0])
        msg = "Invalid DNA byte in sequence {}: '{}'"
        raise ValueError(msg.format(header, bad_character))
    return masked
    
    
class transition(object):
    """
    Mutate Genomic sequence using transitions only.
    :param seq: Original Genomic Sequence.
    :param threshold: probability of Transition.
    :return: Mutated Sequence.
    """

    def __init__(self, threshold):
        self.threshold = threshold
        
    def __call__(self, seq):
    
        A, C, G, T, N = ord('A'), ord('C'), ord('G'), ord('T'), ord('N')

        x = np.random.random(len(seq))
        index = np.where(x < self.threshold)[0]
        #print(index)
        mutations = {A:G, G:A, T:C, C:T, N:N}

        for i in index:
            #print(chr(seq[i]), '->', chr(mutations[seq[i]]))
            seq[i] = mutations[mutations.get(seq[i], N)]

class Random_N(object):
    """
    Mutate Genomic sequence using transitions only.
    :param seq: Original Genomic Sequence.
    :param threshold: probability of Transition.
    :return: Mutated Sequence.
    """

    def __init__(self, n_bp):
        self.n_bp = n_bp
        
    def __call__(self, seq):
    
        N = ord('N')

        index = np.random.randint(0, len(seq), self.n_bp)
        for i in index:
            seq[i] = N


class transversion(object):
    """
    Mutate Genomic sequence (in binary) using transversions only.
    :param seq: Original Genomic Sequence.
    :param threshold: Probability of Transversion.
    :return: Mutated Sequence.
    """
    def __init__(self, threshold):
        self.threshold = threshold
        
    def __call__(self, seq):
        A, C, G, T, N = ord('A'), ord('C'), ord('G'), ord('T'), ord('N')

        x = np.random.random(len(seq))
        index = np.where(x < self.threshold)[0]
        #print(index)

        mutations = {A:[T,C],  G:[T,C], T:[A,G], C:[A,G], N:[N]}

        for i in index:
            seq[i] = random.choice(mutations.get(seq[i], [N]))            
            
class transition_transversion(object):
    """
    Mutate Genomic sequence (in binary) using both transitions 
    and transversions.
    :param seq: Original Genomic Sequence.
    :param threshold: Probability of Transversion.
    :return: Mutated Sequence.
    """
    def __init__(self, threshold_1, threshold_2):
        self.tf1 = transition(threshold_1)
        self.tf2 = transversion(threshold_2)
        
    def __call__(self, seq):

        self.tf1(seq)
        self.tf2(seq)
    
def SummaryFasta(fname, GT_file=None):
    lines = list()
    seq_id = ""
    names, lengths = [], []
    ground_truth = None
    cluster_dis = None

    if GT_file: 
        ground_truth = []
        df = pd.read_csv(GT_file, sep='\t')
        GT_dict = dict(zip(df.sequence_id, df.cluster_id))
        cluster_dis = df['cluster_id'].value_counts().to_dict()

    for line in open(fname, "rb"):

        if line.startswith(b'#'):
            pass

        elif line.startswith(b'>'):
            if seq_id != "":
                seq = bytearray().join(lines)

                if (GT_file and not seq_id in GT_dict):
                    raise ValueError('Check GT for sequence {}'.format(seq_id))

                seq = check_sequence(seq_id, seq)
                names.append(seq_id)
                lengths.append(len(seq))

                if GT_file:
                    ground_truth.append(GT_dict[seq_id])

                lines = []
                seq_id = line[1:-1].decode()  # Modify this according to your labels. 
            
            seq_id = line[1:-1].decode()
        
        else:
            lines += [line.strip()]
  
    if (GT_file and not seq_id in GT_dict):
        raise ValueError('Check GT for sequence {}'.format(seq_id))

    seq = bytearray().join(lines)
    seq = check_sequence(seq_id, seq)
    names.append(seq_id) 
    lengths.append(len(seq))

    if GT_file:
        ground_truth.append(GT_dict[seq_id])

    return names, lengths, ground_truth, cluster_dis


def reverse_complement(x, k):
    numbits = 2*k  
    mask = 0xAAAAAAAA
    x = ((x >> 1) & (mask>>1)) | ((x<< 1) & mask)
    x = (1 << numbits) - 1 - x
    rev = 0

    size = 2**numbits-1
    while(size > 0):
        rev <<= 1
        if x & 1 == 1:
            rev ^= 1
        x >>=1
        size >>= 1

    return rev

def kmer_rev_comp(kmer_counts, k):
    index=[]
    for kmer in range(4**k):
        revcomp = reverse_complement(kmer,k)

        # Only look at canonical kmers - this makes no difference
        if kmer <= revcomp:
            index.append(kmer)
            kmer_counts[kmer] += kmer_counts[revcomp]
            kmer_counts[kmer] *= 0.5
        #else:
        #    kmer_counts[kmer] = kmer_counts[revcomp]

    return kmer_counts[index]


def kmersFasta(fname, k=6, transform=None, reduce=False):
    lines = list()
    seq_id = ""
    names, kmers = [], []

    for line in open(fname, "rb"):
        if line.startswith(b'#'):
            pass

        elif line.startswith(b'>'):
            if seq_id != "":
                seq = bytearray().join(lines)
                names.append(seq_id)  
                            
                if transform:
                    transform(seq)
                    
                counts = np.ones(4**k, dtype=np.int32)
                kmer_counts(seq, k, counts)
                #cgr(seq, k, counts)

                if reduce:
                    counts = kmer_rev_comp(counts,k)


                kmers.append(counts / np.sum(counts))
                
                lines = []
                seq_id = line[1:-1].decode()  # Modify this according to your labels.  
            seq_id = line[1:-1].decode()
        
        else:
            lines += [line.strip()]
  
    seq = bytearray().join(lines)
    names.append(seq_id)
    if transform:
        transform(seq)
        
    counts = np.ones(4**k, dtype=np.int32)
    kmer_counts(seq, k, counts)
    #cgr(seq, k, counts)
    if reduce:
        counts = kmer_rev_comp(counts,k)
    kmers.append(counts / np.sum(counts))
    
    #if reduce:
    #    K_file = np.load(open(f'kernels/kernel{k}.npz','rb'))
    #    KERNEL = K_file['arr_0']
    #    return names, np.dot(np.array(kmers), KERNEL)
        
    return names, np.array(kmers)

def cgrFasta(fname, k=6, transform=None):
    lines = list()
    seq_id = ""
    names, kmers = [], []

    for line in open(fname, "rb"):
        if line.startswith(b'#'):
            pass

        elif line.startswith(b'>'):
            if seq_id != "":
                seq = bytearray().join(lines)
                names.append(seq_id)  
                            
                if transform:
                    transform(seq)
                    
                counts = np.ones(4**k, dtype=np.int32)
                #kmer_counts(seq, k, counts)
                cgr(seq, k, counts)
                kmers.append(counts / np.sum(counts))
                
                lines = []
                seq_id = line[1:-1].decode()  # Modify this according to your labels.  
            seq_id = line[1:-1].decode()
        
        else:
            lines += [line.strip()]
  
    seq = bytearray().join(lines)
    names.append(seq_id)
    if transform:
        transform(seq)
        
    counts = np.ones(4**k, dtype=np.int32)
    #kmer_counts(seq, k, counts)
    cgr(seq, k, counts)
    kmers.append(counts / np.sum(counts))
    return names, np.array(kmers)

 
import time 
def AugmentFasta(sequence_file, n_mimics, k=6, reduce=False):

    train_features = []
    start = time.time()

    # Compute Features and save original data for testing.
    #sys.stdout.write(f'\r............computing augmentations (0/{n_mimics})................')
    #sys.stdout.flush()
    
    _, t_norm = kmersFasta(sequence_file, k=k, transform=transition_transversion(1e-2, 0.5e-2), reduce=reduce)
    t_norm.resize(t_norm.shape[0],1,t_norm.shape[1])
    
    
    #sys.stdout.write(f'\r............computing augmentations (1/{n_mimics})................')
    #sys.stdout.flush()
    _, t_mutated = kmersFasta(sequence_file, k=k, transform=transition(1e-2), reduce=reduce)
    t_mutated.resize(t_mutated.shape[0], 1, t_mutated.shape[1])
    train_features.extend(np.concatenate((t_norm, t_mutated), axis=1))
    
    #sys.stdout.write(f'\r............computing augmentations (2/{n_mimics})................')
    #sys.stdout.flush()
    _, t_mutated = kmersFasta(sequence_file, k=k, transform=transversion(0.5e-2), reduce=reduce)
    t_mutated.resize(t_mutated.shape[0], 1, t_mutated.shape[1])
    train_features.extend(np.concatenate((t_norm, t_mutated), axis=1))
    
    for j in range(n_mimics-2):
        #sys.stdout.write(f'\r............computing augmentations ({3+j}/{n_mimics})................')
        #sys.stdout.flush()
        _, t_mutated = kmersFasta(sequence_file, k=k, transform=Random_N(20), reduce=reduce)
        t_mutated.resize(t_mutated.shape[0], 1, t_mutated.shape[1])
        train_features.extend(np.concatenate((t_norm, t_mutated), axis=1)) 

    x_train = np.array(train_features).astype('float32')
    x_test = np.reshape(t_norm, (-1, t_norm.shape[-1])).astype('float32')
    #print("\n Elapsed Time:", time.time()-start)

    # scaling the data.
    scaler = StandardScaler()
    scaler.fit(x_test)

    x_train_1 = scaler.transform(x_train[:, 0, :])
    x_train_2 = scaler.transform(x_train[:, 1, :])
    x_test = scaler.transform(x_test)

    x_train[:, 0, :] = x_train_1
    x_train[:, 1, :] = x_train_2
    
    return x_train

class AugmentedDataset(Dataset):
    """ 
    Dataset creation directly from fasta file.
    """

    def __init__(self, data):
        """
        : data: (numpy array) of size [:,2,:]
        """
        self.data = data

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        
        sample = {'true': self.data[idx, 0, :], 'modified': self.data[idx, 1, :]}  #<--- We can enforce the prediction of same vector
        return sample

class SequenceDataset(Dataset):
    """ Dataset creation directly from fasta file"""
    
    def __init__(self, fasta_file, k=6, transform=None, GT_file=None, reduce=False):
        """ Args:
            fasta_file (string): Path to te fasta file
            transform (callable, optional): Optional transform to be applied on a 
                                            sequence. Function computing the mimics 
        """
        self.names, self.lengths, self.GT, self.cluster_dis = SummaryFasta(fasta_file, GT_file)
        _, self.kmers = kmersFasta(fasta_file, k, transform, reduce=reduce)

        # scaling the data.
        scaler = StandardScaler()
        self.kmers = scaler.fit_transform(self.kmers)

        
    def __len__(self):
        return len(self.lengths)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
            
        if self.GT:           
            sample = {'kmer':self.kmers[idx,:],'name':self.names[idx], 'cluster_id':self.GT[idx]}
        else:
            sample = {'kmer':self.kmers[idx,:],'name':self.names[idx]}
            
        return sample

def create_dataloader(sequence_file, n_mimics, k=6, batch_size=512, GT_file=None, reduce=False):

    train_data = AugmentFasta(sequence_file, n_mimics, k=k, reduce=reduce)
    training_set = AugmentedDataset(train_data)
    return DataLoader(training_set, batch_size=batch_size, shuffle=True, num_workers=0) #num_workers=4


#-------------------- The following couple of functions are for plotting and saving the point cloud during training-----

def get_coord(probs, num_classes):
    # computes coordinate for 1 sample based on probability distribution over c
  
    coords_total = np.zeros(2, dtype=np.float32)
    probs_sum = probs.sum()

    fst_angle = 0.

    for c in range(num_classes):
        # compute x, y coordinates
        coords = np.ones(2) * 2 * np.pi * (float(c) / num_classes) + fst_angle
        coords[0] = np.sin(coords[0])
        coords[1] = np.cos(coords[1])
        coords_total += (probs[c] / probs_sum) * coords
    return coords_total
  
def PlotPolygon(predictions, num_classes, ax, title):
    hues = np.linspace(0.0, 1.0, num_classes + 1)[0:-1]  # ignore last one
    all_colours = [list((np.array(hsv_to_rgb(hue, 0.8, 0.8)) * 255.)
                   .astype(np.uint8)) for hue in hues]
    
    coordinates = []
    colors = []
    
    #add the points of perfect classification:
    for c in range(num_classes): 
         # compute x, y coordinates
         coords = np.ones(2) * 2 * np.pi * (float(c) / num_classes) 
         coords[0] = np.sin(coords[0])
         coords[1] = np.cos(coords[1])

         #colour = (np.array(all_colours[0][c])).astype(np.uint8)
         colour = (np.zeros(3).astype(np.uint8)) 
         coordinates.append(coords)
         colors.append(colour/255)

    for i in range(predictions.shape[0]):

        coord = get_coord(predictions[i, :], num_classes)
        render_c = np.argmax(predictions[i,:])
        #print(predictions[i, :], render_c)
        #render_c = y_true[i]
        colour = (np.array(all_colours[render_c])).astype(np.uint8)
        coordinates.append(coord)
        colors.append(colour/255)

    coordinates = np.array(coordinates)
    colors = np.array(colors)
    ax.scatter(coordinates[:,0], coordinates[:,1], c = colors, marker = '.', linewidths=0.5)

    ax.axis('equal')
    ax.axis('off')
    ax.set_title('Training Progress (%s)' %(title))
    # plt.savefig(path)

def cluster_acc(y_true, y_pred):
    """
    Calculate clustering accuracy. Require scikit-learn installed
    :param y_true: true labels, numpy.array with shape `(n_samples,)`
    :param y_pred: predicted labels, numpy.array with shape `(n_samples,)`
    :return:  accuracy, in [0,1]
    """

    y_true = y_true.astype(np.int64)
    assert y_pred.size == y_true.size
    D = max(y_pred.max(), y_true.max()) + 1
    w = np.zeros((D, D), dtype=np.int64)

    for i in range(y_pred.size):
        w[y_pred[i], y_true[i]] += 1

    ind = linear_sum_assignment(w.max() - w)
    ind = np.asarray(ind)
    ind = np.transpose(ind)
    return ind, sum([w[i, j] for i, j in ind]) * 1.0 / y_pred.size

def generate_csv_mapping(cm, target_names):
    result_dict = {}
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):  
        result_dict[(target_names[i], j)] = cm[i, j]
    pandas_dict = {
        "Predicted": [],
        "Actual": [],
        "Count": []
    }
    for (target1, target2), count in result_dict.items():
        pandas_dict["Predicted"].append(target2)
        pandas_dict["Actual"].append(target1)
        pandas_dict["Count"].append(count)
    
    result_pd = pd.DataFrame(pandas_dict)
    result_pd.to_csv('confusion_matrix.csv')

def plot_confusion_matrix(cm,
                          target_names,
                          pairs = None,
                          title='Confusion matrix',
                          cmap=None,
                          normalize=False,
                          ax=None):

    
    if not isinstance(pairs, np.ndarray):
        accuracy = np.trace(cm)
    else:
        accuracy = 0
        for i, j in pairs:
            accuracy += cm[j][i]
    accuracy /= float(np.sum(cm))
    misclass = 1 - accuracy

    if cmap is None:
        cmap = plt.get_cmap('Blues')

    ax.imshow(cm, interpolation='nearest', cmap=cmap)
    ax.set_title(title)

    if target_names is not None:
        tick_marks = np.arange(len(target_names))
        ax.set_xticks(tick_marks)
        ax.set_xticklabels([''] * len(target_names))  # , rotation=45)
        ax.set_yticks(tick_marks)
        ax.set_yticklabels(target_names)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    thresh = cm.max() / 1.5 if normalize else cm.max() / 2
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        if normalize:
            ax.text(j, i, "{:0.3f}".format(cm[i, j]),
                    horizontalalignment="center",
                    color="white" if cm[i, j] > thresh else "black")
        else:
            ax.text(j, i, "{:,}".format(cm[i, j]),
                    horizontalalignment="center",
                    color="white" if cm[i, j] > thresh else "black")

    ax.set_ylabel('True label')
    ax.set_xlabel('Predicted label\naccuracy={:0.4f}; misclass={:0.4f}'.format(accuracy, misclass))


from sklearn.metrics.pairwise import euclidean_distances
def label_features(predictions, n_clusters):
    """
    Clustering Ensemble 
    """
    cluster = KMeans(n_clusters=n_clusters, init='k-means++')
    features = np.zeros((predictions.shape[1], predictions.shape[0]*n_clusters))

    for i in range(predictions.shape[0]):
        y_pred = predictions[i]
        for j in range(n_clusters):
            features[y_pred==j, i*n_clusters + j] = 1.

    _sums = np.sum(features, axis=0) / predictions.shape[1]
        
    y = cluster.fit_predict(features - _sums)
    D = 1.0 / euclidean_distances(features - _sums, cluster.cluster_centers_, squared=True)
    D **= 1.0 / (2 - 1)
    D /= np.sum(D, axis=1)[:, np.newaxis]
    fuzzy_labels_ = D
    
    return np.array(y), fuzzy_labels_.max(axis=1)



def compute_results(y_pred, data, y_true=None):
    d = {}

    d['Davies-Boulding'] = metrics.davies_bouldin_score(data, y_pred)
    d['Silhouette-Score'] = metrics.silhouette_score(data, y_pred)

    if not y_true is None:
        d['NMI'] = metrics.adjusted_mutual_info_score(y_true, y_pred)
        d['ARI'] = metrics.adjusted_rand_score(y_true, y_pred)
        d['Homogeneity'] = metrics.homogeneity_score(y_true, y_pred)
        d['Completeness'] = metrics.completeness_score(y_true, y_pred)

        ind, acc = cluster_acc(y_true, y_pred)
        d['ACC'] = acc
    
        return d, ind

    return d, None



        
from itertools import permutations, product
import sys

def modified_Hungarian(test):
    rows, cols = linear_sum_assignment(test)
    new_cols = np.array(list(set(range(test.shape[1])).difference(cols)))

    if new_cols.shape[0] != 0:
        minval = sys.maxsize

        for i in permutations(new_cols, len(new_cols)):
            for j in product(rows, repeat=len(new_cols)):
                _sum = sum(test[j,i])
                if _sum < minval:
                    good_assignment = (j,i)

                minval = min(minval, _sum)

        full_rows = np.array(list(rows) + list(good_assignment[0]))
        full_cols = np.array(list(cols) + list(good_assignment[1]))
    else:
        full_rows, full_cols = rows, cols

    return full_rows, full_cols

def modified_cluster_acc(y_true, y_pred):
    """
    Calculate clustering accuracy with more predicted clusters than true labels
    :param y_true: true labels, numpy.array with shape `(n_samples,)`
    :param y_pred: predicted labels, numpy.array with shape `(n_samples,)`
    :return:  accuracy, in [0,1]
    """

    y_true = y_true.astype(np.int64)
    w = np.zeros((y_true.max() + 1, y_pred.max() + 1), dtype=np.int64)

    for i in range(y_pred.size):
        w[y_true[i], y_pred[i]] += 1
    
    #print(w)

    ind = modified_Hungarian(w.max() - w)
    ind = np.asarray(ind).T
    return ind, sum([w[i, j] for i, j in ind]) * 1.0 / y_pred.size


__version__ = (0, 1, 5)

from idelucs.utils import check_sequence, SummaryFasta, reverse_complement, kmer_rev_comp, kmersFasta, cgrFasta, cluster_acc, compute_results, SequenceDataset, PlotPolygon
from idelucs.kmers import kmer_counts,cgr
from idelucs.models import IID_model
from idelucs.LossFunctions import IID_loss, info_nce_loss
from idelucs.utils_GUI import define_ToolTips


__all__ = [
    "utils",
    "kmers",
    "models",
    "cluster",
    "LossFunctions",
    "utils_GUI",
]


import sys
import torch
import random
import numpy as np
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader

sys.path.append('src/')
sys.path.append('../src/')

from .LossFunctions import IID_loss, info_nce_loss
from .PytorchUtils import NetLinear, myNet
from .ResNet import ResNet18
from .utils import SequenceDataset, create_dataloader

# Random Seeds for reproducibility.
torch.manual_seed(0)
torch.cuda.manual_seed(0)
np.random.seed(0)
random.seed(0)

global dtype
global EPS

dtype = torch.FloatTensor
EPS = sys.float_info.epsilon

device = 'cpu'
if torch.cuda.is_available():
    device = 'cuda'
    dtype = torch.cuda.FloatTensor



def weights_init(m):
    """
    Kaiming initialization of the weights
    :param m: Layer
    :return:
    """
    if isinstance(m, nn.Linear):
        torch.nn.init.kaiming_normal_(m.weight)
        torch.nn.init.zeros_(m.bias)

class IID_model():
    def __init__(self, args: dict):

        self.sequence_file = args['sequence_file']
        self.GT_file = args['GT_file']

        self.n_clusters = args['n_clusters']
        self.k = args['k']
        
        if args['model_size'] == 'linear':
            self.n_features = 4**self.k
            self.net = NetLinear(self.n_features, args['n_clusters'])
            self.reduce = False
            
        elif args['model_size'] == 'small':
            if self.k % 2 == 0: n_in = (4**self.k + 4**(self.k//2))//2 
            else: n_in = (4**self.k)//2
            #d = {4: 135, 5: 511, 6: 2079}
            self.n_features = n_in
            self.net = myNet(self.n_features, args['n_clusters'])
            self.reduce = True
            
        elif args['model_size'] == 'full':
            self.net = ResNet18(1, args['n_clusters'])
        else:
            raise ValueError("Invalid Model Type")
        
        self.net.apply(weights_init)
        self.net.to(device)
        self.epoch = 0
        self.EPS = sys.float_info.epsilon
        
        self.n_mimics = args['n_mimics']
        self.batch_sz = args['batch_sz']
        self.optimizer = args['optimizer']
        self.l = args['lambda'] 
        self.lr = args['lr']
        self.weight = args['weight']
        self.schedule = args['scheduler']
        self.mutate = True

        if self.optimizer == 'RMSprop': 
            self.optimizer = optim.RMSprop(self.net.parameters(), lr=self.lr, weight_decay=0.01)
        elif self.optimizer == 'SGD':
            self.optimizer = optim.SGD(self.net.parameters(), lr=self.lr, weight_decay=0.01, momentum=0.9)
        elif self.optimizer == 'Adam':
            self.optimizer = optim.Adam(self.net.parameters(), lr=self.lr)
        else:
            raise ValueError("Optimizer not supported")
        
        if self.schedule == 'Plateau':
            self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, 'min')
        elif self.schedule == 'Triangle':
            self.scheduler = optim.lr_scheduler.CyclicLR(self.optimizer, base_lr=0.001, max_lr=0.1,step_size_up=5, mode="triangular2")
        
    def build_dataloader(self):
        #Data Files
        data_path = self.sequence_file
        GT_file = self.GT_file
                
        self.dataloader = create_dataloader(data_path, 
                                             self.n_mimics, 
                                             k=self.k, 
                                             batch_size=self.batch_sz, 
                                             GT_file=GT_file,
                                             reduce=self.reduce)

    def contrastive_training_epoch(self):
        self.net.train()
        running_loss = 0.0

        for i_batch, sample_batched in enumerate(self.dataloader):
            sample = sample_batched['true'].view(-1, 1, self.n_features).type(dtype)
            modified_sample = sample_batched['modified'].view(-1, 1, self.n_features).type(dtype)
            
            # zero the gradients
            self.optimizer.zero_grad()

            # forward + backward + optimize
            z1, h1 = self.net(sample)
            z2, h2 = self.net(modified_sample)

            loss = (1-self.weight)*info_nce_loss(h1, h2, 0.85) + (self.weight)*IID_loss(z1, z2, lamb=self.l)
            
            loss.backward()
            self.optimizer.step()

            running_loss += loss
            
        running_loss /= i_batch

        if self.schedule == 'Plateau':
            self.scheduler.step(running_loss)
        elif self.schedule == 'Triangle':
            self.scheduler.step()
        self.epoch += 1

        return running_loss.item()

    def predict(self, data=None):
        
        n_features = self.n_features
        test_dataset = SequenceDataset(self.sequence_file, k=self.k, transform=None, GT_file=self.GT_file, reduce=self.reduce)
        test_dataloader = DataLoader(test_dataset, 
                             batch_size=self.batch_sz,
                             shuffle=False,
                             num_workers=0,
                             drop_last=False)
        y_pred = []
        probabilities = []
        latent = []

        with torch.no_grad():
            self.net.eval()
            
            for test in test_dataloader:
                
                kmers = test['kmer'].view(-1, 1, self.n_features).type(dtype)
                outputs, logits = self.net(kmers)
                probs,  predicted = torch.max(outputs, 1)

                #Extend our list with predictions and groud truth
                y_pred.extend(predicted.cpu().tolist())
                probabilities.extend(probs.cpu().tolist())
                latent.extend(logits.cpu().tolist())
        
        return np.array(y_pred), np.array(probabilities), np.array(latent)
        #return np.array(y_pred), np.array(probabilities), kmers

    def calculate_probs(self, data=None):
        
        n_features = self.n_features
        test_dataset = SequenceDataset(self.sequence_file, k=self.k, transform=None, GT_file=self.GT_file, reduce=self.reduce)
        test_dataloader = DataLoader(test_dataset, 
                             batch_size=self.batch_sz,
                             shuffle=False,
                             num_workers=0,
                             drop_last=False)
 
        probabilities = []
        with torch.no_grad():
            self.net.eval()
            for test in test_dataloader:
                kmers = test['kmer'].view(-1, 1, n_features).type(dtype)

                #calculate the prediction by running through the network
                outputs, logits = self.net(kmers)
                probabilities.extend(outputs.cpu().tolist())

        return np.array(probabilities)


import torch 
import torch.nn as nn

class block(nn.Module):
    def __init__(self, in_channels, out_channels, identity_downsample=None, stride=1):
        super (block, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1)
        self.bn = nn.BatchNorm2d(out_channels)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU()

        self.identity_downsample = identity_downsample
        self.stride = stride

    def forward(self, x):
        identity =  x.clone()

        x = self.conv1(x)
        x = self.bn(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn2(x)

        if self.identity_downsample is not None:
            identity = self.identity_downsample(identity)
            
        x += identity
        x = self.relu(x)

        return x


class ResNet(nn.Module):
    def __init__(self, block, layers, image_channels, num_classes):
        super(ResNet,self).__init__()
        self.in_channels = 16
        self.conv1 = nn.Conv2d(image_channels, 16, kernel_size=5, stride=1, padding=2)
        self.bn = nn.BatchNorm2d(16)
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer_1 = self.make_layer(block, layers[0], 16, 1)
        self.layer_2 = self.make_layer(block, layers[1], 32, 2)
        self.in_channels = 32
        self.layer_3 = self.make_layer(block, layers[2], 64, 2)
        self.in_channels = 64
        self.layer_4 = self.make_layer(block, layers[3], 64, 2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(64, num_classes)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer_1(x)
        x = self.layer_2(x)
        x = self.layer_3(x)
        x = self.layer_4(x)
        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)
        
        return self.softmax(x), x  


    def make_layer(self, block, num_residual_blocks, intermediate_channels, stride):
        identity_downsample = None
        layers = []

        if stride != 1:
            identity_downsample = nn.Sequential(nn.Conv2d(self.in_channels, 
                                                          intermediate_channels, 
                                                          kernel_size=1,
                                                          stride=stride),
                                                nn.BatchNorm2d(intermediate_channels))
                                                

        layers.append(block(self.in_channels, intermediate_channels, identity_downsample, stride))
        
        for i in range(num_residual_blocks - 1):
            layers.append(block(intermediate_channels, intermediate_channels))

        return nn.Sequential(*layers)



class ResNet_k_5(nn.Module):
    def __init__(self, block, layers, image_channels, num_classes):
        super(ResNet_k_5,self).__init__()
        self.in_channels = 16
        self.conv1 = nn.Conv2d(image_channels, 16, kernel_size=5, stride=1, padding=2)
        self.bn = nn.BatchNorm2d(16)
        self.relu = nn.ReLU()
        self.layer_1 = self.make_layer(block, layers[0], 16, 1)
        self.layer_2 = self.make_layer(block, layers[1], 32, 2)
        self.in_channels = 32
        self.layer_3 = self.make_layer(block, layers[2], 64, 2)
        self.in_channels = 64
        self.layer_4 = self.make_layer(block, layers[3], 64, 2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(64, num_classes)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn(x)
        x = self.relu(x)
        x = self.layer_1(x)
        x = self.layer_2(x)
        x = self.layer_3(x)
        x = self.layer_4(x)
        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)
        
        return self.softmax(x), x  


    def make_layer(self, block, num_residual_blocks, intermediate_channels, stride):
        identity_downsample = None
        layers = []

        if stride != 1:
            identity_downsample = nn.Sequential(nn.Conv2d(self.in_channels, 
                                                          intermediate_channels, 
                                                          kernel_size=1,
                                                          stride=stride),
                                                nn.BatchNorm2d(intermediate_channels))
                                                

        layers.append(block(self.in_channels, intermediate_channels, identity_downsample, stride))
        
        for i in range(num_residual_blocks - 1):
            layers.append(block(intermediate_channels, intermediate_channels))

        return nn.Sequential(*layers)


def ResNet18(img_channel=3, num_classes=50):
    return ResNet(block, [2,2,2,2], img_channel, num_classes)

def ResNet18_k_5(img_channel=3, num_classes=50):
    return ResNet_k_5(block, [2,2,2,2], img_channel, num_classes)


class ConvNet( nn.Module):
    def __init__(self, img_channel = 1, num_classes=10):
        super(ConvNet,self).__init__()
        self.num_classes = num_classes
        self.layers = nn.Sequential(
                nn.Conv2d(img_channel, 16, 5, 1, 2),
                nn.BatchNorm2d(16),
                nn.ReLU(),
                nn.AvgPool2d(3,2,1),
                nn.Conv2d(16,32,3,1,1),
                nn.BatchNorm2d(32),
                nn.ReLU(),
                nn.MaxPool2d(3,2,1),
                nn.Conv2d(32,64,3,1,1),
                nn.BatchNorm2d(64),
                nn.ReLU(),
                nn.MaxPool2d(3,2,1),
                nn.Conv2d(64,64,3,1,1),
                nn.AdaptiveAvgPool2d((1,1))
        )

        self.fc = nn.Linear(64, num_classes)
        self.softmax = nn.Softmax(dim=1)
    
    def forward(self, x):
        latent = self.layers(x)
        latent = latent.reshape(latent.shape[0], -1)
        out = self.softmax(self.fc(latent))

        return out, latent 


class ConvNet_k_5( nn.Module):
    def __init__(self, img_channel = 1, num_classes=10):
        super(ConvNet_k_5,self).__init__()
        self.num_classes = num_classes
        self.layers = nn.Sequential(
                nn.Conv2d(img_channel, 16, 5, 1, 2),
                nn.BatchNorm2d(16),
                nn.ReLU(),
                nn.Conv2d(16,32,3,1,1),
                nn.BatchNorm2d(32),
                nn.ReLU(),
                nn.MaxPool2d(3,2,1),
                nn.Conv2d(32,64,3,1,1),
                nn.BatchNorm2d(64),
                nn.ReLU(),
                nn.MaxPool2d(3,2,1),
                nn.Conv2d(64,64,3,1,1),
                nn.AdaptiveAvgPool2d((1,1))
        )

        self.fc = nn.Linear(64, num_classes)
        self.softmax = nn.Softmax(dim=1)
    
    def forward(self, x):
        latent = self.layers(x)
        latent = latent.reshape(latent.shape[0], -1)
        out = self.softmax(self.fc(latent))

        return out, latent 


def test(device):
    net = ResNet18_k_5(img_channel=1)
    x = torch.rand(2,1, 64, 64)

    #net = ConvNet(img_channel=1)
    #x = torch.rand(2,1, 64, 64)


    print(net)
    net.to(device)
    y = net(x)
    print(y[0].shape)

#test('cpu')




from setuptools import setup, find_packages, Extension

setup(
        include_package_data=True,
        packages=find_packages(),
        
        classifiers= [
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        ext_modules=[Extension(name="ideluc.kmers", sources=["idelucs/kmers.pyx"])]
    )


import sys
import torch

import torch.nn as nn
import torch.nn.functional as F


global label_dtype
global dtype

dtype = torch.FloatTensor
label_dtype = torch.LongTensor

device = 'cpu'
if torch.cuda.is_available():
    device = 'cuda'
    dtype = torch.cuda.FloatTensor
    label_dtype = torch.cuda.LongTensor

def IID_loss(x_out, x_tf_out, lamb=1.0, EPS=sys.float_info.epsilon):
    """
    Implementation of the IID loss function found in the paper:
    "Invariant Information Clustering for Unsupervised Image 
    Classification and Segmentation"
    The function can be found originally in https://github.com/xu-ji/IIC
    """
    
    _, k = x_out.size()
    p_i_j = compute_joint(x_out, x_tf_out)
    assert (p_i_j.size() == (k, k))

    p_i = p_i_j.sum(dim=1).view(k, 1).expand(k, k).clone()
    p_j = p_i_j.sum(dim=0).view(1, k).expand(k, k).clone()  # but should be same, symmetric

    # avoid NaN losses. Effect will get cancelled out by p_i_j tiny anyway
    p_i_j[(p_i_j < EPS).data] = EPS
    p_j[(p_j < EPS).data] = EPS
    p_i[(p_i < EPS).data] = EPS

    loss = - p_i_j * (torch.log(p_i_j)
                      - lamb * torch.log(p_j)
                      - lamb * torch.log(p_i))

    loss = loss.sum()

    return loss


def compute_joint(x_out, x_tf_out):
    """
    Produces variable that requires grad (since args require grad)
    The function can be found originally in https://github.com/xu-ji/IIC
    """
    bn, k = x_out.size()
    assert (x_tf_out.size(0) == bn and x_tf_out.size(1) == k)

    p_i_j = x_out.unsqueeze(2) * x_tf_out.unsqueeze(1)  # bn, k, k
    p_i_j = p_i_j.sum(dim=0)  # k, k
    p_i_j = (p_i_j + p_i_j.t()) / 2.  # symmetrise
    p_i_j = p_i_j / p_i_j.sum()  # normalise

    return p_i_j


def info_nce_loss(z1,z2, temperature):
    """Info-NCE los function.
        Introduced by "A Simple Framework for Contrastive 
        Learning of Visual Representations" by T. Chen,
        S. Kornblith, M. Norouzi, and Geoffrey Hinton 
        (https://arxiv.org/abs/2002.05709)
    """
    N = z1.shape[0]
    criterion = nn.CrossEntropyLoss()
    features = torch.cat((z1, z2), 0).float()
    labels = torch.cat([torch.arange(N) for i in range(2)], dim=0)
    labels = (labels.unsqueeze(0) == labels.unsqueeze(1)).float()

    features = F.normalize(features, dim=1)

    similarity_matrix = torch.matmul(features, features.T)

    # discard the main diagonal from both: labels and similarities matrix
    mask = torch.eye(labels.shape[0], dtype=torch.bool)
    labels = labels[~mask].view(labels.shape[0], -1)
    similarity_matrix = similarity_matrix[~mask].view(similarity_matrix.shape[0], -1)

    # select and combine multiple positives
    positives = similarity_matrix[labels.bool()].view(labels.shape[0], -1)

    # select only the negatives the negatives
    negatives = similarity_matrix[~labels.bool()].view(similarity_matrix.shape[0], -1)

    logits = torch.cat([positives, negatives], dim=1)
    labels = torch.zeros(logits.shape[0]).type(label_dtype)

    logits = logits / temperature
    loss = criterion(logits, labels)
    return loss



#!/usr/bin/env python3

import sys
import pandas as pd
import os
import numpy as np
import matplotlib.pyplot as plt


from idelucs.utils import SummaryFasta, plot_confusion_matrix, \
                      label_features, compute_results

from idelucs import models

import argparse
import torch
import sys
import time

import csv
#-------------------------------
import hdbscan

def weights_init(m):
    """
    Kaiming initialization of the weights
    :param m: Layer
    :return:
    """
    if isinstance(m, torch.nn.Linear):
        torch.nn.init.kaiming_normal_(m.weight)
        torch.nn.init.zeros_(m.bias)
        
def save_results_in_file(dataset_name, model_name, model_parameters, results, time, memory, file_name):
    # Create the file if it does not exitst and write first file with column names
    with open(file_name, mode='a', newline='') as file:
        writer = csv.writer(file, delimiter='\t')
        if file.tell() == 0:
            writer.writerow(['Dataset', 'Model', 'Parameters'] + list(results.keys()) + ['Time', 'Memory'])

        # Write the new row
        row = [dataset_name, model_name, model_parameters] + list(results.values()) + [time, memory]
        writer.writerow(row)


def run(args):

    start_time = time.time()
    now = time.asctime()
    time_stamp = now.split(' ')
    hour = now.split(' ')[3]
    time_stamp[3] = '-'.join(hour.split(':'))
    
    time_stamp = '_'.join(time_stamp[1:-1])
    
    folder_name = os.getcwd()
    folder_name = f'{folder_name}/Results'
    
    use_hdbscan = False

    if not os.path.isdir(folder_name):
        os.mkdir(folder_name)

    results_folder = f'{folder_name}/{(args["sequence_file"]).split("/")[-1].split(".")[0]}'
    
    if not os.path.isdir(f'{results_folder}'):
        os.mkdir(f'{results_folder}')
    
    os.mkdir(f'{results_folder}/{time_stamp}')

    if args['n_clusters'] == 0:
        args['n_clusters'], use_hdbscan = 200, True
        
    model = models.IID_model(args)
    model.names, model.lengths, model.GT, model.cluster_dis = SummaryFasta(model.sequence_file,
                                                                           model.GT_file)
    
    if use_hdbscan:
        clusterer = hdbscan.HDBSCAN(min_cluster_size=len(model.names)//100 + 1, gen_min_span_tree=True, prediction_data=True)
    
    print(model.cluster_dis)
    stats = {"n_seq": len(model.lengths),
            "min_len": np.min(model.lengths),
            "max_len": np.max(model.lengths),
            "avg_len": np.mean(model.lengths)}

    print( f'No. Sequences: \t {stats["n_seq"]:,}')
    print(f'Min. Length: \t {stats["min_len"]:,}')
    print(f'Max. Length: \t {stats["max_len"]:,}')
    print(f'Avg. Length: \t {round(stats["avg_len"],2):,}')
    model.build_dataloader()
    predictions = []
    min_loss = np.inf
    
    training_figure, display_training = plt.subplots(nrows=1, ncols=1)
    display_training.grid(True)
    display_training.set_title("Learning Curves")
    display_training.set_xlabel("Epoch")
    display_training.set_ylabel("Training Loss")


    for voter in range(args['n_voters']):
        sys.stdout.write(f"\r........... Training Model ({voter+1}/{ args['n_voters']})................")
        sys.stdout.flush()
        model.net.apply(weights_init)
        model.epoch = 0
        model_loss = []
        model_min_loss=np.inf

        for i in range(args['n_epochs']):
            loss = model.contrastive_training_epoch()
            model_min_loss = min(model_min_loss, loss)
            model_loss.append(loss)

        length = len(model.names)
        y_pred, probabilities, latent = model.predict()

        if model_min_loss < min_loss:
            model_latent = latent
        
        if not use_hdbscan:
            y_pred = y_pred.astype(np.int32)
            
            d = {}
            count = 0
            for i in range(y_pred.shape[0]):
                if y_pred[i] in d:
                    y_pred[i] = d[y_pred[i]]
                else:
                    d[y_pred[i]] = count
                    y_pred[i] = count
                    count += 1
            predictions.append(y_pred)

        display_training.plot(model_loss, label=f'Model {voter+1}')
        display_training.axes.legend(loc=1)
        training_figure.savefig(f'{results_folder}/{time_stamp}/training_plots.jpg')
                                                                

    if not use_hdbscan:
        y_pred, probabilities = label_features(np.array(predictions), args['n_clusters'])
        
    else:
        clusterer.fit(latent )
        y_pred = clusterer.labels_ + 1
        probabilities = clusterer.probabilities_
        args['n_clusters'] = np.max(y_pred) + 1
        

    #--------------------- Computing and Saving the Results
    sys.stdout.write(f"\r........... Computing Results ................")
    sys.stdout.flush()
    
    if args['GT_file'] != None:
        unique_labels = list(np.unique(model.GT))
        numClasses = len(unique_labels)
        y = np.array(list(map(lambda x: unique_labels.index(x), model.GT)))
        results, ind = compute_results(y_pred, latent, y)

        d = {}
        for i, j in ind:
            d[i] = j
        
        if -1 in y_pred:
            d[-1] = 0
        
        w = np.zeros((numClasses, max(max(y_pred) + 1, max(y) + 1)), dtype=np.int64)
        clustered = np.zeros_like(y, dtype=bool)
        for i in range(y.shape[0]):
            w[y[i], d[y_pred[i]]] += 1

            if y[i] == d[y_pred[i]]:
                clustered[i] = True
         
        if args['n_clusters'] < 16:
            fig, new_ax = plt.subplots( nrows=1, ncols=1)  # create figure & 1 axis
            plot_confusion_matrix(w, unique_labels, ax=new_ax, normalize=False)
            fig.savefig(f'{results_folder}/{time_stamp}/contingency_matrix.jpg')
        else:
            # generating a csv mapping
            #generate_csv_mapping(w, unique_labels)
            w_df = pd.DataFrame(w)
            w_df.index = unique_labels
            w_df.to_csv(f'{results_folder}/{time_stamp}/contingency_matrix.tsv',sep='\t')

    else:
        results, ind = compute_results(y_pred, latent)
        #clustered = (probabilities >= 0.9)
        clustered = (probabilities >= 0.0)
    
    print(results)
    sys.stdout.write(f"\r........ Saving Results ..............")
    sys.stdout.flush()
    
    
    
    dataset_name = args['sequence_file'].split('/')[-1]
    del args['sequence_file']
    del args['GT_file']
    

    _time = (time.time() - start_time) #running_info.ru_utime + running_info.ru_stime
    hour = _time // 3600
    minutes = (_time  - (3600 * hour)) // 60
    seconds = _time - (hour * 3600) - (minutes * 60)
    memory = "" #(running_info.ru_maxrss/1e6)
    print("") #running_info)
    
    names = np.array(model.names)
    data = np.concatenate((names[:,np.newaxis],
                              y_pred[:,np.newaxis],
                              probabilities[:,np.newaxis]), axis=1)

    df = pd.DataFrame(data, columns=['sequence_id','assignment','confidence_score'])
    df.to_csv(f'{results_folder}/{time_stamp}/assignments.tsv',sep='\t')

    df=pd.Series(results, name='Value')
    df.to_csv(f'{results_folder}/{time_stamp}/metrics.tsv',sep='\t')
    
    save_results_in_file(dataset_name, "iDeLUCS", args, results, 
                         f'{hour}:{minutes}:{seconds}', 
                         memory, f'{os.getcwd()}/ALL_RESULTS.tsv')

    # ------------------------------- Computing and Saving the Representations
    if args['plot']:
        sys.stdout.write(f"\r........... Computing Representations................")
        sys.stdout.flush()

        import umap
        n_samples = len(model.names)
        mask = np.full(n_samples, True)
        MAX_SAMPLES = 1000
        if n_samples > MAX_SAMPLES :
            mask = np.full(n_samples, False)
            mask[:MAX_SAMPLES] = True
            np.random.shuffle(mask)

        embedding = umap.UMAP(random_state=42).fit_transform(latent)
        fig, ax = plt.subplots(nrows=1, ncols=1) 
        ax.set_title("Representation of the Latent Space")
        ax.set_xlabel("UMAP 1")
        ax.set_ylabel("UMAP 2")

        ax.scatter(embedding[mask & ~clustered, 0],
                   embedding[mask & ~clustered, 1],
                   c = np.atleast_2d([0.5, 0.5, 0.5]),
                   s=1,
                   alpha=0.5)
        ax.scatter(embedding[mask & clustered, 0],
                  embedding[mask & clustered, 1],
                  c=y[mask & clustered],  #y_pred[mask & clustered]
                  s=1,
                  cmap='Spectral')
        fig.savefig(f'{results_folder}/{time_stamp}/learned_representation.jpg', dpi=150)
    
def main():
    parser= argparse.ArgumentParser()
    parser.add_argument('--sequence_file', action='store',type=str)
    parser.add_argument('--n_clusters', action='store',type=int,default=0,
                        help='Expected or maximum number of clusters to find. \n'
                            'It should be equal or greater than n_true_clusters \n'
                            'when GT is provided. \n'
                            'NOTE: Use 0 for automatically finding fine-grained \n'
                            'clusters')
    parser.add_argument('--n_epochs', action='store',type=int,default=100,
                        help='Number of training epochs. An epoch is defined \n' 
                             'as a training iteration over all the training pairs.')
    parser.add_argument('--n_mimics', action='store',type=int,default=3, 
                        help='Number of data augmentations per sequence \n' 
                            'that will be considered during training.')
    parser.add_argument('--batch_sz', action='store',type=int,default=256)
    parser.add_argument('--GT_file', action='store',type=str,default=None)
    parser.add_argument('--k', action='store',type=int,default=6, help="k-mer length")
    parser.add_argument('--optimizer', action='store',type=str,default="RMSprop")
    parser.add_argument('--scheduler', action='store',type=str,default="None")

    parser.add_argument('--weight', action='store',type=float,default=0.25,
                        help='Relative importance of the contrastive objective on \n'
                            'the final loss. Use a higher value when low intra- \n'
                            'cluster distance is expected and a lower value when \n'
                            'high intra-cluster variability is expected')
    
    parser.add_argument('--lambda', action='store',type=float,default=2.8,
                        help='Hyperparameter to control cluster balance. \n'
                            'Use lambda: 1.2 when unbalanced clusters are expected \n'
                            'Use lambda: 2.8 when perfectly balanced clusters are expected \n')
    parser.add_argument('--lr', action='store',type=float,default=1e-3, help="Learning Rate")
    parser.add_argument('--n_voters', action='store',type=int, default=5, help="Number of Voters")
    parser.add_argument('--model_size', action='store',type=str,default="linear", help="Selection of 'conv', 'small', 'linear' or 'full'")
    parser.add_argument('--plot', action='store',type=bool, default=False, help="Set to True to plot the final output representation")
    
    args = vars(parser.parse_args())

    print("\nTraining Parameters:")
    for key in args:
        print(f'{key} \t -> {args[key]}')
    
    run(args)


if __name__ == '__main__':
    main()
